"""A configuration option for click."""

from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Optional, Sequence, Type, Union
from urllib.parse import urlparse

import httpx
from click import Context, Option
from click.decorators import FC, _param_memo  # noqa: PLC2701

from bumpversion.exceptions import BadInputError, BumpVersionError
from bumpversion.ui import get_indented_logger

logger = get_indented_logger(__name__)

BoolOrStr = Union[bool, str]
StrSequence = Sequence[str]


class ConfigOption(Option):
    """A configuration option for click."""

    def __init__(
        self,
        param_decls: Optional[StrSequence] = None,
        show_default: Optional[BoolOrStr] = None,
        allow_from_autoenv: bool = True,
        help: Optional[str] = None,
        show_envvar: bool = False,
        **attrs,
    ):
        param_decls = param_decls or ("--config", "-C")
        multiple = False
        count = False
        hidden = False
        show_choices = True
        prompt = False
        confirmation_prompt = False
        is_flag = None
        flag_value = None
        prompt_required = True
        hide_input = False
        type_ = str
        meta_var = "PATH_OR_URL"

        super().__init__(
            param_decls=param_decls,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            prompt_required=prompt_required,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            metavar=meta_var,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            type=type_,
            help=help,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            **attrs,
        )

    def process_value(self, ctx: Context, value: Any) -> Optional[Path]:
        """Process the value of the option."""
        value = super().process_value(ctx, value)
        return resolve_conf_location(value) if value else None


def config_option(*param_decls: str, cls: Optional[Type[ConfigOption]] = None, **attrs: Any) -> Callable[[FC], FC]:
    """
    Attaches a ConfigOption to the command.

    All positional arguments are passed as parameter declarations to `ConfigOption`.

    All keyword arguments are forwarded unchanged (except ``cls``). This is equivalent to creating a
    `ConfigOption` instance manually and attaching it to the `Command.params` list.

    For the default option class, refer to `ConfigOption` and `Parameter` for descriptions of parameters.

    Args:
        *param_decls: Passed as positional arguments to the constructor of `cls`.
        cls: the option class to instantiate.  This defaults to `ConfigOption`.
        **attrs: Passed as keyword arguments to the constructor of `cls`.

    Returns:
        A decorated function.
    """
    if cls is None:  # pragma: no-coverage
        cls = ConfigOption

    def decorator(f: FC) -> FC:
        _param_memo(f, cls(param_decls, **attrs))
        return f

    return decorator


def resolve_conf_location(url_or_path: str) -> Path:
    """Resolve a URL or path.

    The path is considered a URL if it is parseable as such and starts with ``http://`` or ``https://``.

    Args:
        url_or_path: The URL or path to resolve.

    Raises:
        BumpVersionError: if the file does not exist.

    Returns:
        The contents of the location.
    """
    parsed_url = urlparse(url_or_path)

    if parsed_url.scheme in ("http", "https"):
        return download_url(url_or_path)

    path = Path(url_or_path)
    if not path.exists():
        raise BumpVersionError(f"'{path}' does not exist.")
    return path


def download_url(url: str) -> Path:
    """
    Download the contents of a URL.

    Args:
        url: The URL to download

    Returns:
        The Path to the downloaded file.

    Raises:
        BadInputError: if there is a problem downloading the URL
    """
    logger.debug(f"Downloading configuration from URL: {url}")
    filename = get_file_name_from_url(url)
    suffix = Path(filename).suffix

    try:
        resp = httpx.get(url, follow_redirects=True, timeout=1)
        resp.raise_for_status()
        with NamedTemporaryFile(mode="w", delete=False, encoding="utf-8", suffix=suffix) as tmp:
            tmp.write(resp.text)
        return Path(tmp.name)
    except httpx.RequestError as e:
        raise BadInputError(f"Unable to download configuration from URL: {url}") from e
    except httpx.HTTPStatusError as e:
        msg = f"Error response {e.response.status_code} while requesting {url}."
        raise BadInputError(msg) from e


def get_file_name_from_url(url: str) -> str:
    """
    Extracts the file name from a URL.

    Args:
        url: The URL to extract the file name from.

    Returns:
        The file name from the URL, or an empty string if there is no file name.
    """
    parsed_url = urlparse(url)

    return parsed_url.path.split("/")[-1]



"""Testing fixtures for Pytest."""

import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Tuple

import pytest
from click.testing import CliRunner

from bumpversion.config import DEFAULTS, Config
from bumpversion.scm.models import SCMConfig
from bumpversion.versioning.models import Version
from bumpversion.versioning.version_config import VersionConfig


@pytest.fixture
def tests_path() -> Path:
    """Return the path containing the tests."""
    return Path(__file__).parent


@pytest.fixture
def fixtures_path(tests_path: Path) -> Path:
    """Return the path containing the testing fixtures."""
    return tests_path / "fixtures"


@contextmanager
def inside_dir(dirpath: Path) -> Generator:
    """
    Temporarily switch to a specific directory.

    Args:
        dirpath: Path of the directory to switch to
    """
    import os

    old_path = os.getcwd()
    try:
        os.chdir(dirpath)
        yield
    finally:
        os.chdir(old_path)


def get_config_data(overrides: dict) -> Tuple[Config, VersionConfig, Version]:
    """Get the configuration, version_config and version."""
    from bumpversion import config

    conf = config.get_configuration(config_file="missing", **overrides)
    version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
    version = version_config.parse(conf.current_version)

    return conf, version_config, version


def get_semver(version: str) -> Version:
    """Get a semantic version from a string."""
    _, _, version = get_config_data({"current_version": version})
    return version


@pytest.fixture
def git_repo(tmp_path: Path) -> Path:
    """Generate a simple temporary git repo and return the path."""
    subprocess.run(
        ["git", "init", "--initial-branch=main"], cwd=tmp_path, check=True, capture_output=True  # noqa: S607
    )
    subprocess.run(
        ["git", "config", "--local", "--type", "bool", "commit.gpgsign", "false"],  # noqa: S607
        cwd=tmp_path,
        check=True,
        capture_output=True,
    )
    subprocess.run(
        ["git", "config", "--local", "--type", "bool", "tag.gpgsign", "false"],  # noqa: S607
        cwd=tmp_path,
        check=True,
        capture_output=True,
    )
    return tmp_path


@pytest.fixture
def hg_repo(tmp_path: Path) -> Path:
    """Generate a simple temporary mercurial repo and return the path."""
    subprocess.run(["hg", "init"], cwd=tmp_path, check=True, capture_output=True)  # noqa: S603, S607
    return tmp_path


@pytest.fixture
def runner() -> CliRunner:
    """Return a CliRunner instance."""
    return CliRunner()


@pytest.fixture
def scm_config() -> SCMConfig:
    """Dummy SCMConfig fixture."""
    return SCMConfig(
        tag=DEFAULTS["tag"],
        sign_tags=DEFAULTS["sign_tags"],
        tag_name=DEFAULTS["tag_name"],
        allow_dirty=DEFAULTS["allow_dirty"],
        commit=DEFAULTS["commit"],
        message=DEFAULTS["message"],
        moveable_tags=DEFAULTS["moveable_tags"],
        parse_pattern=DEFAULTS["parse"],
        tag_message=DEFAULTS["tag_message"],
        commit_args=DEFAULTS["commit_args"],
    )



"""File processing tests."""

import shutil
from datetime import datetime
from difflib import context_diff
from pathlib import Path
from textwrap import dedent

import pytest
import tomlkit
from pytest import param

from bumpversion import bump, config, exceptions, files
from bumpversion.config.models import FileChange
from bumpversion.context import get_context
from bumpversion.exceptions import VersionNotFoundError
from bumpversion.versioning.version_config import VersionConfig
from tests.conftest import get_config_data, inside_dir


def test_do_bump_uses_correct_serialization_for_tag_and_commit_messages(git_repo: Path, fixtures_path: Path, caplog):
    """The tag and commit messages should use the root configured serialization."""
    import subprocess
    import logging

    caplog.set_level(logging.INFO)

    csharp_path = fixtures_path / "csharp"
    dst_path: Path = shutil.copytree(csharp_path, git_repo / "csharp")
    subprocess.check_call(["git", "add", "."], cwd=git_repo)
    subprocess.check_call(["git", "commit", "-m", "Initial commit"], cwd=git_repo)
    subprocess.check_call(["git", "tag", "3.1.0-rc+build.1031"], cwd=git_repo)

    with inside_dir(dst_path):
        config_file = dst_path.joinpath(".bumpversion.toml")
        conf = config.get_configuration(config_file=config_file)
        bump.do_bump("patch", None, conf, config_file, dry_run=True)

    assert "Bump version: 3.1.0-rc+build.1031 -> 3.1.1" in caplog.text


class TestModifyFiles:
    """Tests for the modify_files function."""

    def test_raises_correct_missing_version_string(self, tmp_path: Path, fixtures_path: Path):
        """When a file is missing the version string, the error should indicate the correct serialization missing."""
        csharp_path = fixtures_path / "csharp"
        dst_path: Path = shutil.copytree(csharp_path, tmp_path / "csharp")

        with inside_dir(dst_path):
            conf = config.get_configuration(config_file=dst_path.joinpath(".bumpversion.toml"))
            version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
            current_version = version_config.parse(conf.current_version)
            dst_path.joinpath("Version.csv").write_text("1;3-1;0;rc;build.1031", encoding="utf-8")

            major_version = current_version.bump("patch")
            ctx = get_context(conf)

            configured_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

            with pytest.raises(VersionNotFoundError) as e:
                files.modify_files(configured_files, current_version, major_version, ctx)
                assert e.message.startswith("Did not find '1;3;1;0;rc;build.1031'")

    def test_non_matching_search_does_not_modify_file(self, tmp_path: Path):
        """If the search string is not found in the file, do nothing."""
        changelog_path = tmp_path / "CHANGELOG.md"
        changelog_content = dedent(
            """
        # Unreleased

        * bullet point A prep for 1.0.3

        # Release v'older' (2019-09-17)

        * bullet point B
        """
        )
        changelog_path.write_text(changelog_content, encoding="utf-8")

        overrides = {
            "current_version": "1.0.3",
            "files": [
                {
                    "filename": str(changelog_path),
                    "search": "Not-yet-released",
                    "replace": "Release v{new_version} ({now:%Y-%m-%d})",
                }
            ],
        }
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        configured_files = files.resolve_file_config(conf.files, version_config)
        with pytest.raises(exceptions.VersionNotFoundError, match="Did not find 'Not-yet-released' in file:"):
            files.modify_files(configured_files, current_version, new_version, get_context(conf))

        assert changelog_path.read_text() == changelog_content

    @pytest.mark.parametrize(
        ["global_value", "file_value", "should_raise"],
        [
            param(True, True, False, id="ignore global and file"),
            param(True, False, True, id="ignore global only"),
            param(False, True, False, id="ignore file only"),
            param(False, False, True, id="ignore none"),
        ],
    )
    def test_ignore_missing_version_ignores_correctly(
        self, global_value: bool, file_value: bool, should_raise: bool, tmp_path: Path
    ) -> None:
        """Evaluate the various configurations of ignore_missing_version."""
        # Arrange
        version_path = tmp_path / Path("VERSION")
        version_path.write_text("1.2.3", encoding="utf-8")

        overrides = {
            "current_version": "1.2.5",
            "ignore_missing_version": global_value,
            "files": [{"filename": str(version_path), "ignore_missing_version": file_value}],
        }
        with inside_dir(tmp_path):
            conf, version_config, current_version = get_config_data(overrides)
            new_version = current_version.bump("patch")
            cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

            # Act
            if should_raise:
                with pytest.raises(VersionNotFoundError):
                    files.modify_files(cfg_files, current_version, new_version, get_context(conf))
            else:
                files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        assert version_path.read_text() == "1.2.3"


class TestModifyFilesRegex:
    """Tests of regular expression search and replace in modify_files."""

    def test_regex_search_string_is_replaced(self, tmp_path: Path) -> None:
        """A regex search string is found and replaced."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("Release: 1234-56-78 '1.2.3'", encoding="utf-8")

        overrides = {
            "regex": True,
            "current_version": "1.2.3",
            "search": r"Release: \d{{4}}-\d{{2}}-\d{{2}} '{current_version}'",
            "replace": r"Release {now:%Y-%m-%d} '{new_version}'",
            "files": [{"filename": str(version_path)}],
        }
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        now = datetime.now().isoformat()[:10]
        assert version_path.read_text() == f"Release {now} '1.2.4'"

    def test_escaped_chars_are_respected(self, tmp_path: Path) -> None:
        """A search that uses escaped characters is respected in a regex search and replaces."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("## [Release] 1.2.3 1234-56-78", encoding="utf-8")

        overrides = {
            "regex": True,
            "current_version": "1.2.3",
            "search": r"## \[Release\] {current_version} \d{{4}}-\d{{2}}-\d{{2}}",
            "replace": r"## [Release] {new_version} {now:%Y-%m-%d}",
            "files": [{"filename": str(version_path)}],
        }
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        now = datetime.now().isoformat()[:10]
        assert version_path.read_text() == f"## [Release] 1.2.4 {now}"

    def test_search_in_toml(self, tmp_path: Path, fixtures_path: Path) -> None:
        """Tests how-to doc 'update-a-date.md'."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("__date__ = '1234-56-78'\n__version__ = '1.2.3'", encoding="utf-8")
        config_path = tmp_path / ".bumpversion.toml"
        shutil.copyfile(fixtures_path / "replace-date-config.toml", config_path)
        with inside_dir(tmp_path):
            conf = config.get_configuration(config_file=config_path)
            version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
            current_version = version_config.parse(conf.current_version)

            new_version = current_version.bump("patch")
            cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

            # Act
            files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        now = datetime.now().isoformat()[:10]
        assert version_path.read_text() == f"__date__ = '{now}'\n__version__ = '1.2.4'"

    def test_caret_matches_beginning_of_line(self, tmp_path: Path, fixtures_path: Path) -> None:
        """A search that uses a caret to indicate the beginning of the line works correctly."""
        # Arrange
        config_path = tmp_path / ".bumpversion.toml"
        thingy_path = tmp_path / "thingy.yaml"
        shutil.copyfile(fixtures_path / "regex_with_caret.yaml", thingy_path)
        shutil.copyfile(fixtures_path / "regex_with_caret_config.toml", config_path)

        conf = config.get_configuration(config_file=config_path)
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        new_version = current_version.bump("patch")

        with inside_dir(tmp_path):
            cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

            # Act
            files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        assert (
            thingy_path.read_text()
            == "version: 1.0.1\ndependencies:\n- name: kube-prometheus-stack\n  version: 1.0.0\n"
        )

    def test_bad_regex_falls_back_to_string_search(self, tmp_path: Path, caplog) -> None:
        """A search string not meant to be a regex is still found and replaced."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("Score: A+ ( '1.2.3'", encoding="utf-8")

        overrides = {
            "regex": True,
            "current_version": "1.2.3",
            "search": r"Score: A+ ( '{current_version}'",
            "replace": r"Score: A+ ( '{new_version}'",
            "files": [{"filename": str(version_path)}],
        }
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        cfg_files = [files.ConfiguredFile(file_cfg, version_config) for file_cfg in conf.files]

        # Act
        files.modify_files(cfg_files, current_version, new_version, get_context(conf))

        # Assert
        assert version_path.read_text() == "Score: A+ ( '1.2.4'"
        assert "Invalid regex" in caplog.text


class TestDataFileUpdater:
    """Tests for the DataFileUpdater class."""

    def test_update_file_does_not_modify_non_toml_files(self, tmp_path: Path) -> None:
        """A non-TOML file is not modified."""
        # Arrange
        version_path = tmp_path / "VERSION"
        version_path.write_text("1.2.3", encoding="utf-8")

        overrides = {"current_version": "1.2.3", "files": [{"filename": str(version_path)}]}
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        datafile_config = FileChange(
            filename=str(version_path),
            key_path="",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        assert version_path.read_text() == "1.2.3"

    def test_update_replaces_key(self, tmp_path: Path, fixtures_path: Path) -> None:
        """A key specific key is replaced and nothing else is touched."""
        # Arrange
        config_path = tmp_path / "pyproject.toml"
        fixture_path = fixtures_path / "partial_version_strings.toml"
        shutil.copy(fixture_path, config_path)

        contents_before = config_path.read_text()
        conf = config.get_configuration(config_file=config_path, files=[{"filename": str(config_path)}])
        version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
        current_version = version_config.parse(conf.current_version)
        new_version = current_version.bump("minor")
        datafile_config = FileChange(
            filename=str(config_path),
            key_path="tool.bumpversion.current_version",
            search=conf.search,
            replace=conf.replace,
            regex=conf.regex,
            ignore_missing_version=conf.ignore_missing_version,
            ignore_missing_file=conf.ignore_missing_files,
            serialize=conf.serialize,
            parse=conf.parse,
        )

        # Act
        files.DataFileUpdater(datafile_config, version_config.part_configs).update_file(
            current_version, new_version, get_context(conf)
        )

        # Assert
        contents_after = config_path.read_text()
        toml_data = tomlkit.parse(config_path.read_text()).unwrap()
        actual_difference = list(
            context_diff(
                contents_before.splitlines(),
                contents_after.splitlines(),
                fromfile="before",
                tofile="after",
                n=0,
                lineterm="",
            )
        )
        expected_difference = [
            "*** before",
            "--- after",
            "***************",
            "*** 28 ****",
            '! current_version = "0.0.2"',
            "--- 28 ----",
            '! current_version = "0.1.0"',
        ]
        assert actual_difference == expected_difference
        assert toml_data["tool"]["pdm"]["dev-dependencies"]["lint"] == ["ruff==0.0.292"]
        assert toml_data["tool"]["bumpversion"]["current_version"] == "0.1.0"



"""Tests for the ConfiguredFile class."""

import logging
import os
import shutil
from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent

import pytest

from bumpversion import files
from bumpversion.context import get_context
from bumpversion.files import ConfiguredFile, FileChange
from bumpversion.ui import get_indented_logger, setup_logging
from bumpversion.versioning.models import VersionComponentSpec
from bumpversion.versioning.version_config import VersionConfig
from tests.conftest import get_config_data, inside_dir


def setup_configured_file(filepath):
    """Sets up a configured file for testing."""
    overrides = {
        "current_version": "1.2.3",
        "files": [{"filename": str(filepath)}],
    }
    conf, version_config, _current_version = get_config_data(overrides)
    assert len(conf.files_to_modify) == 1
    assert conf.files_to_modify[0].filename == str(filepath)
    return ConfiguredFile(conf.files_to_modify[0], version_config)


class TestConfiguredFile:
    """Tests for the ConfiguredFile class."""

    class TestClassCreation:
        """Tests for the creation of the ConfiguredFile class."""

        def test_file_change_is_identical_to_input(self):
            """Test that the file change is identical to the input but a different object."""
            # Arrange
            change = FileChange(
                filename="boobar.txt",
                search="dashes: {current_version}",
                replace="dashes: {new_version}",
                parse=r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                serialize=("{major}-{minor}-{patch}",),
                regex=True,
                ignore_missing_version=False,
                ignore_missing_file=False,
            )
            version_config = VersionConfig(
                parse="(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
                serialize=("{major}.{minor}.{patch}",),
                search="{current_version}",
                replace="{new_version}",
                part_configs={
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                },
            )

            # Act
            cfg_file = ConfiguredFile(change, version_config)

            # Assert
            assert cfg_file.file_change == change
            assert cfg_file.file_change is not change

        def test_version_config_uses_file_change_attrs(self):
            """Test that the version config uses the file change attributes and the original part configs."""
            # Arrange
            change = FileChange(
                filename="boobar.txt",
                search="dashes: {current_version}",
                replace="dashes: {new_version}",
                parse=r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                serialize=("{major}-{minor}-{patch}",),
                regex=True,
                ignore_missing_version=False,
                ignore_missing_file=False,
            )
            version_config = VersionConfig(
                parse="(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
                serialize=("{major}.{minor}.{patch}",),
                search="{current_version}",
                replace="{new_version}",
                part_configs={
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                },
            )
            expected = VersionConfig(
                parse=change.parse,
                search=change.search,
                replace=change.replace,
                serialize=change.serialize,
                part_configs=version_config.part_configs,
            )

            # Act
            cfg_file = ConfiguredFile(change, version_config)

            # Assert
            assert cfg_file.version_config == expected

    class TestGetFileContents:
        """Tests of the get_file_content method."""

        def test_returns_contents_of_existing_file(self, tmp_path: Path) -> None:
            """The get_file_content method returns the contents of an existing file."""
            # Arrange
            filepath = tmp_path / "file.md"
            expected_data = "My current version is 1.2.3"
            filepath.write_text(expected_data, encoding="utf-8")
            configured_file = setup_configured_file(filepath)

            # Assert
            assert configured_file.get_file_contents() == expected_data

        def test_raises_exception_if_file_not_found(self, tmp_path: Path) -> None:
            """The get_file_content method raises an exception if the file does not exist."""
            filepath = tmp_path / "file.md"
            configured_file = setup_configured_file(filepath)
            # Act
            with pytest.raises(FileNotFoundError):
                configured_file.get_file_contents()

    class TestMakeFileChange:
        """Tests of the make_file_change function."""

        def test_raises_exception_if_file_not_found(self, tmp_path: Path) -> None:
            """The make_file_change method raises an exception if the file does not exist."""
            # Arrange
            filepath = tmp_path / "file.md"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath)}],
            }
            with inside_dir(tmp_path):
                conf, version_config, current_version = get_config_data(overrides)
            configured_file = ConfiguredFile(conf.files_to_modify[0], version_config)
            new_version = current_version.bump("patch")
            ctx = get_context(conf)

            # Act
            with pytest.raises(FileNotFoundError):
                configured_file.make_file_change(current_version, new_version, ctx, dry_run=True)

        def test_logs_missing_file_when_ignore_missing_file_set(self, tmp_path: Path, caplog) -> None:
            """The make_file_change method logs missing file when ignore_missing_file set to True."""
            self._setup_logging(caplog)
            filepath = tmp_path / "file.md"
            overrides = {
                "current_version": "1.2.3",
                "files": [{"filename": str(filepath), "ignore_missing_file": True}],
            }
            with inside_dir(tmp_path):
                conf, version_config, current_version = get_config_data(overrides)
                configured_file = ConfiguredFile(conf.files_to_modify[0], version_config)
                new_version = current_version.bump("patch")
                ctx = get_context(conf)

                # Act
                configured_file.make_file_change(current_version, new_version, ctx, dry_run=True)

            # Assert
            logs = caplog.messages
            assert logs[0] == "Reading configuration"
            assert logs[1] == "  Configuration file not found: missing."
            assert logs[2] == f"\nFile {filepath}: replace `{{current_version}}` with `{{new_version}}`"
            assert logs[3] == "  File not found, but ignoring"

        def test_dedents_properly_when_file_does_not_contain_pattern(
            self, fixtures_path: Path, tmp_path, caplog
        ) -> None:
            """The make_file_change method dedents the logging context when it does not contain a pattern."""
            self._setup_logging(caplog)
            overrides = {
                "current_version": "1.2.3",
                "files": [{"glob": "glob/**/*.txt", "ignore_missing_file": True, "ignore_missing_version": True}],
            }
            shutil.copytree(fixtures_path / "glob", tmp_path / "glob")

            with inside_dir(tmp_path):
                conf, version_config, current_version = get_config_data(overrides)
                configured_file1 = ConfiguredFile(conf.files_to_modify[0], version_config)
                configured_file2 = ConfiguredFile(conf.files_to_modify[1], version_config)
                configured_file3 = ConfiguredFile(conf.files_to_modify[2], version_config)
            new_version = current_version.bump("patch")
            ctx = get_context(conf)

            # Act
            configured_file1.make_file_change(current_version, new_version, ctx, dry_run=True)
            configured_file2.make_file_change(current_version, new_version, ctx, dry_run=True)
            configured_file3.make_file_change(current_version, new_version, ctx, dry_run=True)

            # Assert
            logs = caplog.messages
            assert logs[0] == "Reading configuration"
            assert logs[1] == "  Configuration file not found: missing."
            assert logs[2] == (
                f"\nFile {configured_file1.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )
            assert logs[3] == "  File not found, but ignoring"
            assert logs[4] == (
                f"\nFile {configured_file2.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )
            assert logs[5] == "  File not found, but ignoring"
            assert logs[6] == (
                f"\nFile {configured_file3.file_change.filename}: replace `{{current_version}}` with `{{new_version}}`"
            )
            assert logs[7] == "  File not found, but ignoring"

        def _setup_logging(self, caplog):
            setup_logging(1)
            caplog.set_level(logging.INFO)
            logger = get_indented_logger(__name__)
            logger.reset()

    def test_can_process_single_file_twice(self, tmp_path: Path):
        """
        Verify that a single file "file2" can be processed twice.

        Use two file_ entries, both with a different suffix after the underscore.
        Employ different parse/serialize and search/replace configs to verify correct interpretation.
        """
        # Arrange
        filepath = tmp_path.joinpath("file2")
        filepath.write_text("dots: 0.10.2\ndashes: 0-10-2", encoding="utf-8")
        overrides = {
            "current_version": "0.10.2",
            "files": [
                {
                    "filename": str(filepath),
                    "search": "dots: {current_version}",
                    "replace": "dots: {new_version}",
                },
                {
                    "filename": str(filepath),
                    "search": "dashes: {current_version}",
                    "replace": "dashes: {new_version}",
                    "parse": r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                    "serialize": ["{major}-{minor}-{patch}"],
                },
            ],
        }
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")
        ctx = get_context(conf)

        # Act
        for file_cfg in conf.files:
            cfg_file = files.ConfiguredFile(file_cfg, version_config)
            cfg_file.make_file_change(current_version, new_version, ctx)

        # Assert
        assert len(conf.files) == 2
        assert filepath.read_text() == "dots: 0.10.3\ndashes: 0-10-3"

    def test_can_multi_file_configuration(self, tmp_path: Path):
        full_vers_path = tmp_path / "FULL_VERSION.txt"
        full_vers_path.write_text("1.0.3", encoding="utf-8")
        maj_vers_path = tmp_path / "MAJOR_VERSION.txt"
        maj_vers_path.write_text("1", encoding="utf-8")
        readme_path = tmp_path / "README.txt"
        readme_path.write_text("MyAwesomeSoftware(TM) v1.0", encoding="utf-8")
        build_num_path = tmp_path / "BUILD_NUMBER"
        build_num_path.write_text("1.0.3+joe+38943", encoding="utf-8")

        overrides = {
            "current_version": "1.0.3+joe+38943",
            "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:\+.*)?",
            "serialize": [
                "{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}",
                "{major}.{minor}.{patch}",
            ],
            "files": [
                {
                    "filename": str(full_vers_path),
                },
                {
                    "filename": str(maj_vers_path),
                    "parse": r"\d+",
                    "serialize": ["{major}"],
                },
                {
                    "filename": str(readme_path),
                    "parse": r"(?P<major>\d+)\.(?P<minor>\d+)",
                    "serialize": ["{major}.{minor}"],
                },
                {
                    "filename": str(build_num_path),
                    "serialize": ["{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}"],
                },
            ],
        }
        conf, version_config, current_version = get_config_data(overrides)
        major_version = current_version.bump("major")

        os.environ["BUILD_NUMBER"] = "38944"
        os.environ["USER"] = "bob"
        ctx = get_context(conf)
        del os.environ["BUILD_NUMBER"]
        del os.environ["USER"]

        for file_cfg in conf.files:
            cfg_file = files.ConfiguredFile(file_cfg, version_config)
            cfg_file.make_file_change(current_version, major_version, ctx)

        assert full_vers_path.read_text() == "2.0.0"
        assert maj_vers_path.read_text() == "2"
        assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
        assert build_num_path.read_text() == "2.0.0+bob+38944"

        os.environ["BUILD_NUMBER"] = "38945"
        os.environ["USER"] = "jane"
        conf.current_version = "2.0.0+bob+38944"
        ctx = get_context(conf)
        del os.environ["BUILD_NUMBER"]
        del os.environ["USER"]
        major_version = version_config.parse(conf.current_version)
        major_patch_version = major_version.bump("patch")
        for file_cfg in conf.files:
            cfg_file = files.ConfiguredFile(file_cfg, version_config)
            cfg_file.make_file_change(major_version, major_patch_version, ctx)

        assert full_vers_path.read_text() == "2.0.1"
        assert maj_vers_path.read_text() == "2"
        assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
        assert build_num_path.read_text() == "2.0.1+jane+38945"


def test_make_file_change_search_replace_to_avoid_updating_unconcerned_lines(tmp_path: Path, caplog):
    import logging

    caplog.set_level(logging.DEBUG)
    req_path = tmp_path / "requirements.txt"
    req_path.write_text("Django>=1.5.6,<1.6\nMyProject==1.5.6", encoding="utf-8")

    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_path.write_text(
        dedent(
            """
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
        ),
        encoding="utf-8",
    )

    overrides = {
        "current_version": "1.5.6",
        "files": [
            {
                "filename": str(req_path),
                "search": "MyProject=={current_version}",
                "replace": "MyProject=={new_version}",
            },
            {
                "filename": str(changelog_path),
                "search": "## [Unreleased]",
                "replace": "## [Unreleased]\n\n## [{new_version}] - {utcnow:%Y-%m-%d}",
                "no_regex": True,
            },
        ],
    }

    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("minor")

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    utc_today = datetime.now(timezone.utc).strftime("%Y-%m-%d")
    expected_chglog = dedent(
        f"""
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]

    ## [1.6.0] - {utc_today}
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
    )
    assert req_path.read_text() == "Django>=1.5.6,<1.6\nMyProject==1.6.0"
    assert changelog_path.read_text() == expected_chglog


def test_make_file_change_simple_replacement_in_utf8_file(tmp_path: Path):
    """Changing a file in UTF-8 should not change the non-ASCII characters."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_bytes("Kröt1.3.0".encode("utf-8"))

    overrides = {"current_version": "1.3.0", "files": [{"filename": str(version_path)}]}
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch")

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    # Assert
    out = version_path.read_text()
    assert out == "Kröt1.3.1"


def test_make_file_change_multi_line_search_is_found(tmp_path: Path) -> None:
    """A multiline search string is found and replaced."""
    # Arrange
    alphabet_path = tmp_path / "the_alphabet.txt"
    alphabet_path.write_text("A\nB\nC\n", encoding="utf-8")

    overrides = {
        "current_version": "9.8.7",
        "search": "A\nB\nC",
        "replace": "A\nB\nC\n{new_version}",
        "files": [{"filename": str(alphabet_path)}],
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("major")

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.make_file_change(current_version, new_version, get_context(conf))

    # Assert
    assert alphabet_path.read_text() == "A\nB\nC\n10.0.0\n"
